package tools.descartes.prisma.agent.wildfly.service;
import java.io.File;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;

import org.jboss.logging.Logger;
import org.jboss.msc.service.Service;
import org.jboss.msc.service.ServiceName;
import org.jboss.msc.service.StartContext;
import org.jboss.msc.service.StartException;
import org.jboss.msc.service.StopContext;
import org.jboss.msc.value.InjectedValue;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.TraceRoot;
import tools.descartes.prisma.agent.wildfly.callstack.TraceVisitor;
import tools.descartes.prisma.agent.wildfly.callstack.impl.DefaultComponentRecord;
import tools.descartes.prisma.agent.wildfly.filters.FilterDefinition;
import tools.descartes.prisma.agent.wildfly.filters.FilteredTraceFactory;
import tools.descartes.prisma.agent.wildfly.parameters.ValueCharacterization;


public class MonitoringService implements Service<MonitoringService> {
	
	private static class Printer implements TraceVisitor {
		
		private long timestamp;
		private PrintWriter out;
		
		public void prepare(PrintWriter out) {
			this.out = out;
			timestamp = System.currentTimeMillis();
		}
		
		public void finish() {
			out = null;
		}
		
		@Override
		public void visitComponent(ComponentRecord component) {
		}

		@Override
		public void visitInterface(InterfaceRecord interf) { 
		}
		

		@Override
		public void visitOperation(OperationRecord record) {
		}

		@Override
		public void visitInvocationGroup(InvocationGroupRecord group) {
			out.printf("%d, %s, %s, %s, %s, , , , , %d, %d, %s\n", 
					timestamp, 
					group.getOperation().getInterface().getComponent().getComponentName(),
					group.getOperation().getInterface().getInterfaceName(),
					group.getOperation().getOperationName(),
					group.getName(),
					group.getInvocationCount(),
					group.getExecutionTime(),
					group.getAdditionalInformation()
				);			
		}

		@Override
		public void visitRequiredComponent(InvocationGroupRecord source, ComponentRecord component) { }

		@Override
		public void visitRequiredInterface(InvocationGroupRecord source, InterfaceRecord interf) { }

		@Override
		public void visitRequiredOperation(InvocationGroupRecord source, OperationRecord record) { }

		@Override
		public void visitRequiredInvocationGroup(InvocationGroupRecord source, InvocationGroupRecord group) {
			out.printf("%d, %s, %s, %s, %s, %s, %s, %s, %s, %d, %d, %s\n", 
					timestamp, 
					source.getOperation().getInterface().getComponent().getComponentName(),
					source.getOperation().getInterface().getInterfaceName(),
					source.getOperation().getOperationName(),
					source.getName(),
					group.getOperation().getInterface().getComponent().getComponentName(),
					group.getOperation().getInterface().getInterfaceName(),
					group.getOperation().getOperationName(),
					group.getName(),
					group.getInvocationCount(),
					group.getExecutionTime(),
					group.getAdditionalInformation()
					);
			
		}

	}
	
	private static final Logger log = Logger.getLogger(MonitoringService.class);
	
	public static final ServiceName SERVICE_NAME = ServiceName.JBOSS.append("modelextraction", "monitoring");
	
	public static final ServiceName FILTER_NAME = SERVICE_NAME.append("filter");
	
	public static final ServiceName PARAM_CHARACTERIZATION_NAME = SERVICE_NAME.append("filter");
	
	public static final MonitoringService INSTANCE = new MonitoringService();
	
	private static final long WRITE_INTERVAL = 60000;
	
	private final TraceRoot root = new TraceRoot(new FilteredTraceFactory());
	
	private final ConcurrentHashMap<Long, Long> threads = new ConcurrentHashMap<Long, Long>();
	
	private List<InjectedValue<FilterService>> filters = new CopyOnWriteArrayList<>();
	
	private List<InjectedValue<ParameterCharacterizationService>> paramCharacterizations = new CopyOnWriteArrayList<>();
	
	private Thread OUTPUT = new Thread() {
		public void run() {
			try {
				Printer printer = new Printer();
				
				while(true) {
					Thread.sleep(WRITE_INTERVAL);
					
					try(final PrintWriter out = new PrintWriter(new FileWriter(new File("./operations.csv"), true))) {
						printer.prepare(out);
						root.accept(printer);						
					} catch(Exception ex) {
						log.error("Error writing monitoring data.", ex);
					}
					printer.finish();
				}
			} catch(InterruptedException ex) {
				interrupted();
			}
		};
	};

	@Override
	public MonitoringService getValue() throws IllegalStateException, IllegalArgumentException {
		return this;
	}

	@Override
	public void start(StartContext ctx) throws StartException {
		FilteredTraceFactory factory = (FilteredTraceFactory) root.getTraceFactory();
		for (InjectedValue<FilterService> filter : filters) {
			filter.getValue().addPredicate(factory);
		}
		for (InjectedValue<ParameterCharacterizationService> param : paramCharacterizations) {
			factory.addValueCharacterization(param.getValue().createValueCharacterization());					
		}
		System.err.println("MonitoringService started");		
		OUTPUT.start();	
	}

	@Override
	public void stop(StopContext arg0) {
		OUTPUT.interrupt();
	}
	
	public TraceRoot getTraceRoot() {
		return root;
	}
	
	public long getLastCpuTime(long threadId) {
		Long val = threads.get(threadId);
		return (val == null) ? 0 : val;
	}
	
	public void setLastCpuTime(long threadId, long cpuTimeInNanos) {
		threads.put(threadId, cpuTimeInNanos);
	}
	
	public List<InjectedValue<FilterService>> getFilters() {
		return filters;
	}
	
	public List<InjectedValue<ParameterCharacterizationService>> getParameterCharacterizations() {
		return paramCharacterizations;
	}

}
