package tools.descartes.prisma.core;

import java.util.List;

import org.eclipse.emf.cdo.eresource.CDOResource;
import org.eclipse.emf.cdo.eresource.CDOResourceFolder;
import org.eclipse.emf.cdo.transaction.CDOTransaction;
import org.eclipse.emf.cdo.util.CommitException;
import org.eclipse.emf.cdo.util.ConcurrentAccessException;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.Repository;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.system.System;
import edu.kit.ipd.descartes.mm.applicationlevel.system.SystemFactory;
import edu.kit.ipd.descartes.mm.deployment.Deployment;
import edu.kit.ipd.descartes.mm.deployment.DeploymentFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.DistributedDataCenter;
import edu.kit.ipd.descartes.mm.resourcelandscape.ResourcelandscapeFactory;
import edu.kit.ipd.descartes.mm.usageprofile.UsageProfile;
import edu.kit.ipd.descartes.mm.usageprofile.UsageProfileFactory;
import tools.descartes.prisma.core.commands.CommandContext;
import tools.descartes.prisma.core.commands.ModelCommand;

public class ApplicationModel {
	
	private static final int TRANSACTION_RETRIES = 3;
	
	private final RepositoryConnection connection;
	private final String name;	
	
	public ApplicationModel(RepositoryConnection connection, String name) {
		this.connection = connection;
		this.name = name;
	}
	
	public String getResourceLandscapePath() {
		return name + "/" + name + ".resourcelandscape";
	}
	
	public String getRepositoryPath() {
		return name + "/" + name + ".repository";
	}
	
	public String getSystemPath() {
		return name + "/" + name + ".system";
	}
	
	public String getDeploymentPath() {
		return name + "/" + name + ".deployment";
	}
	
	public String getUsageProfilePath() {
		return name + "/" + name + ".usageprofile";
	}
	
	public void update(List<ModelCommand<?>> commands) {
		boolean successful = false;
		int retries = TRANSACTION_RETRIES;
		do {
			CDOTransaction transaction = connection.startTransaction();
			CommandContext context = new CommandContext(this, transaction);
			for (ModelCommand<?> currentCommand : commands) {
				currentCommand.apply(context);
			}
			try {
				transaction.commit();
				successful = true;
			} catch (ConcurrentAccessException e) {
				transaction.rollback();
				e.printStackTrace();
			} catch (CommitException e) {
				transaction.rollback();
				e.printStackTrace();
			}
			retries--;
		} while(!successful && retries >= 0);
	}
	
	public void create() {
		try {
			CDOTransaction transaction = connection.startTransaction();
			CDOResourceFolder appFolder = transaction.createResourceFolder(name);
			
			CDOResource resourceLandscapeResource = appFolder.addResource(name + ".resourcelandscape");
			DistributedDataCenter dataCenter = ResourcelandscapeFactory.eINSTANCE.createDistributedDataCenter();
			dataCenter.setName(name);
			resourceLandscapeResource.getContents().add(dataCenter);
			
			CDOResource repositoryResource = appFolder.addResource(name + ".repository");
			Repository repository = RepositoryFactory.eINSTANCE.createRepository();
			repository.setName(name);
			repositoryResource.getContents().add(repository);
			
			CDOResource systemResource = appFolder.addResource(name + ".system");
			System system = SystemFactory.eINSTANCE.createSystem();
			system.setName(name);
			systemResource.getContents().add(system);
			
			CDOResource deploymentResource = appFolder.addResource(name + ".deployment");
			Deployment deployment = DeploymentFactory.eINSTANCE.createDeployment();
			deployment.setName(name);
			deployment.setSystem(system);
			deployment.setTargetResourceLandscape(dataCenter);
			deploymentResource.getContents().add(deployment);
			
			CDOResource usageProfileResource = appFolder.addResource(name + ".usageprofile");
			UsageProfile usage = UsageProfileFactory.eINSTANCE.createUsageProfile();
			usage.setName(name);
			usage.setSystem(system);
			usageProfileResource.getContents().add(usage);
			
			transaction.commit();
		} catch(Exception ex) {
			ex.printStackTrace();
		}
	}

}
